// a combined batch processing recon and stat program
#include <unistd.h>
#include <stdio.h>
#include <math.h>
#include <complex.h>
#include <fstream.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>
#include "propfun.h"
#include "byte_swap.h"
#include <string.h>
#define PI      M_PI
#define MAX_S   256
#define NCPUS 8
#define CONTROL "batstat.sim"
#define FALSE   0
#define TRUE    1
#define ST strcmp
#define TS temp_string
#define MAX_FILES       64
#define SMALL_INT -128
#define	BG_TOKEN  0
#define	MG_TOKEN  1
#define FG_TOKEN  2
#define PEAK_TOKEN 3
#define BOOL    int
#define MAX_OPTIONS 2500
#define NBINS 256
#define SMALL_NO 1e-6
#define PLANE         0
#define GAUSSIAN      1
#define DOUGHNUT      2
#define MAGCONST      1
typedef struct {
        float   sx,ssqx;
        float   xbar,sdx;
        float   xmin,xmax;
        int     n;
        int     dropouts;
        } STAT;
 
#define COMPLEX_TYPE double

//prototypes for internal fft's 
void fourn_c(complex<COMPLEX_TYPE>  *,unsigned long *,int,int);
void normalizefft(complex<COMPLEX_TYPE>  *, unsigned long *, int );

//prototypes for functions
void 		add_wavefront(complex<COMPLEX_TYPE> *, int ,int , int ,int ,float,float,float);
void 		filter(complex<COMPLEX_TYPE> * ,int ,int);
void            intext_gaussian(char *,complex<COMPLEX_TYPE> *, int , int ,int ,int ,int ,float );
void 		linedump(char *,char *,complex<COMPLEX_TYPE> *,int ,int,int,int ,float);
float           sd(float, float, float, int);
void            intext  (char *, complex<COMPLEX_TYPE> *, int,int, int, int, int);
void            output(BOOL ,char *, complex<COMPLEX_TYPE> *, int,int);
void		outcom_bin(char  *, complex<COMPLEX_TYPE> *, int,int);
void            outint_bin(char  *, int*, int,int);
float           find_max_int    (complex<COMPLEX_TYPE> *, int);
float           image_psf_size(float,float,float);
void            draw_psf(int,int,int,int,int *,int,int);
void            mask_array(complex<COMPLEX_TYPE> *,complex<COMPLEX_TYPE> *,int *,int,int);
void            outint_from_com(char *, complex<COMPLEX_TYPE> *, int, int);
void            collect_stats(complex<COMPLEX_TYPE> *, int *, int,int ,STAT *,STAT *,STAT * ,STAT *);
float           intensity(complex<COMPLEX_TYPE>);
float           sd(float, float, float, int);
void            draw_mask(char *,int *,float ,float ,int ,int ,float ,int,int );
void            read_ok(char *);
void            collect_histogram(complex<COMPLEX_TYPE> *, int *, int , int *, int *, STAT *, STAT *, char *,float);
void 		checki(int);
void 		init_work(complex<COMPLEX_TYPE> * ,int ,int );
void 		fft(complex<COMPLEX_TYPE> *,int ,int ,int );
void 		prop2d(complex<COMPLEX_TYPE> * , float ,float ,float ,int ,int );
void		outint_txt(char *,int *,int,int );	
inline complex<COMPLEX_TYPE> 	rot(complex<COMPLEX_TYPE> , double );
inline complex<COMPLEX_TYPE> rotf(double ,double ,double );
//-------------------------------------------------------------
//---important****large 2d arrays sothese are going to be padded---
//---so that the y stride is an odd number,
//---------------------------------------------------------

//---global variable for the cray fft routine workspace stuff---------
        //int             ntable,nwork;
        //float           *work,*table;
	float detune=1.00;
	//lazy bastard
	
	//swap byte order for alpha2sun conversion
	BOOL 		swap_floats=FALSE;

main(int argc,char **argv){
//variables
// stream declarations
        ifstream        is,control;
        ofstream        os;
	FILE		*of;
// working arrays
        complex<COMPLEX_TYPE>        *data;
        complex<COMPLEX_TYPE>        *fground,*bground,*pground;
        int            *mask;
        int             *hfg,*hbg;      //histograms

// basic parameters and incidentals
        int             rank=0,phaselevels[MAX_FILES],i,ystride=0,N=0,nx[MAX_FILES],ny[MAX_FILES];
        float           wavelength=633e-9,F[MAX_FILES];
	float		step=1000,psize[MAX_FILES];
        float   	variable=0,dol=.1;      //contains the option value 
	float		scale,tempf;
        int     	offx,offy;  //offsets in recon pixels;
	int 		tempi1,tempi2;
	int		filter_width=10;
// stat variables
        STAT    	pg_stat,fg_stat,mg_stat,bg_stat;
        float   	contrast,efficiency;

// filenames and the like
        char            control_file[MAX_S],temp_string[MAX_S],statfile[MAX_S];
        char            iname[MAX_FILES][MAX_S], phase_files[MAX_FILES][MAX_S];
	char		oname[MAX_FILES][MAX_S],reconfile[MAX_S];
	char		posfile[MAX_FILES][MAX_S];
	float 		var[MAX_FILES],R_curve[MAX_FILES];
	BOOL		phase_on[MAX_FILES],btype[MAX_FILES];
	int 		no_files=0;
// logics
        BOOL            hol_out=FALSE,recon_out=FALSE,recon_file=FALSE,linedump_on=FALSE;
	BOOL		bgout=FALSE,histogram=FALSE,pgout=FALSE,fftout=FALSE,mgout=FALSE,maskout=FALSE;
        BOOL    	psf_set_meters=FALSE,psf_set_pixels=FALSE;
        BOOL    	statfile_set=FALSE,newstart=FALSE;
	BOOL		complex_out=TRUE,narrow_on=FALSE;
	BOOL		spat_filt=FALSE;
// incidentals and counters etc
	int     	ix,iy;
       	float   	x,y,temp_no;
        float   	Ax,Ay;    //size of image psf in meters
        int     	ax,ay;      //1/2 size of image psf in reconstruction array pixels
	int		temp;
//phase variables
        float   	phase,sphase=0,ssqphase=0,maxphase=-1e6,minphase=1e6;
        float   	pbar=-9,psd=-9,lphase=-9,ssqdphase=0,dphasebar=0;
        int     	nph=0;
cout<<"Sim_ap- \n";
cout<<"knows spherical input/byte_swapping/00\n";
cout<<"      gaussian and plane beams\n";
cout<<"Version 4.30 Jan 1999\n";
cout<<"Knows oname output_filname command\n";
cout<<"Knows fftfile command (complex<COMPLEX_TYPE> output)\n";
cout<<"Knows spatial_filter xx (for 0 order xx in pixels)"<<endl;

for(i=0;i<MAX_FILES;i++){
		phase_on[i]=FALSE;
		psize[i]=.00001;
		nx[i]=0;
		ny[i]=0;
		phaselevels[i]=2;
                btype[i]=PLANE;
		F[i]=.1;
		R_curve[i]=0;
	}	



// first check and open control file
// start of the program
	         if (argc > 1) {
                strcpy(control_file, argv[1]);
                }
        else {
                strcpy(control_file, CONTROL);
                }
	read_ok(control_file);
        control.open(control_file);
        if(control.bad()){
		cout<<"ERROR:: unable to open control file "<<control_file<<"\n";
                exit(3);
                }
// readin file and find 'options' bit and then set options
        i=0;
	do{
		i++;
                control>>temp_string;
                       if(ST(TS,"gaussian")==0){
                        for(temp=0;temp<MAX_FILES;temp++){
                                btype[temp]=GAUSSIAN;
                                }
                        }
                if(ST(TS,"output_intensity")==0){
			complex_out=FALSE;
			}
	        if(ST(TS,"input_curvature")==0){
			control>>tempf;
                       for(temp=0;temp<MAX_FILES;temp++){
                                R_curve[temp]=tempf;
                                }
			}
                if(ST(TS,"swap_floats")==0){
                        swap_floats=TRUE;
                        }
		if(ST(TS,"plane")==0){
                        for(temp=0;temp<MAX_FILES;temp++){
                                btype[temp]=PLANE;
                                }
                        }
         	if(ST(TS,"recon_out")==0){
                        recon_out=TRUE;
                        }
		if(ST(TS,"narrow")==0){
                        narrow_on=TRUE;
                        }
                if(ST(TS,"spatial_filter")==0){
			control>>filter_width;
			spat_filt=TRUE;
			}
		if(ST(TS,"wavelength")==0){
                        control>>wavelength;
                        }
                if(ST(TS,"posfile")==0){
			control>>temp_string;
			for(temp=0;temp<MAX_FILES;temp++){
				strcpy(posfile[temp],temp_string);
				}
			}
		if(ST(TS,"hologram")==0){
                        temp=0;
                        do{
                               	control>>temp_string;
				if(ST(TS,"name")==0){
					control>>iname[temp++];
                               		strcpy(oname[temp-1],iname[temp-1]);
					}
				if(ST(TS,"oname")==0){
					control>>oname[temp-1];
					}
				if(ST(TS,"phase_file")==0){
					control>>phase_files[temp-1];
					phase_on[temp-1]=TRUE;
					}
				if(ST(TS,"var")==0||ST(TS,"variable")==0){
					control>>var[temp-1];
					}
                                if(ST(TS,"gaussian")==0){
                                        btype[temp-1]=GAUSSIAN;
                                        }
                                if(ST(TS,"plane")==0){
                                        btype[temp-1]=PLANE;
                                        }
				if(ST(TS,"posfile")==0){
                        		control>>posfile[temp-1];
                                	}
			        if(ST(TS,"psize")==0){
                        		control>>psize[temp-1];
                       		         }
                                if(ST(TS,"pixels")==0){
                                        control>>nx[temp-1]>>ny[temp-1];
                                         }
				if(ST(TS,"phase")==0){
					control>>phaselevels[temp-1];
					}
			        if(ST(TS,"focal_length")==0){
                        		control>>F[temp-1];
                        		}
                		if(ST(TS,"input_curvature")==0){
                        		control>>R_curve[temp-1];
                        	}
			}
                        while(ST(TS,"end")!=0&&temp<MAX_FILES&&!control.eof());
                        no_files=temp;
                        cout<<temp<<" files specified \n";
		}
                if(ST(TS,"focal_length")==0){
                        control>>tempf;
                        for(temp=0;temp<MAX_FILES;temp++){
                                F[temp]=tempf;
                                }
                        }
                if(ST(TS,"reconsize")==0){
                        control>>N;
                        }
                if(ST(TS,"pixels")==0){
                        control>>tempi1>>tempi2;
                        for(temp=0;temp<MAX_FILES;temp++){
                                nx[temp]=tempi1;
				ny[temp]=tempi2;
                                }
                        }
                if(ST(TS,"psize")==0){
                        control>>tempf;
                       	for(temp=0;temp<MAX_FILES;temp++){
                                psize[temp]=tempf;
                                }
                        }
                if(ST(TS,"phase")==0){
                        control>>tempi1;
			for(temp=0;temp<MAX_FILES;temp++){
				phaselevels[temp]=tempi1;
				}
                        }
		if(ST(TS,"detune")==0){
			control>>detune;
		cout<<"Phase detune factor="<<detune*100<<"%\n";}
               if(ST(TS,"bgfile")==0){
                	bgout=TRUE;
                	}
        	if(ST(TS,"histogram")==0){
                	histogram=TRUE;
                	}
		if(ST(TS,"linedump")==0){
			linedump_on=TRUE;
			}
                if(ST(TS,"pgfile")==0){
                        pgout=TRUE;
                        }
        	if(ST(TS,"fftfile")==0){
                	fftout=TRUE;
                	}
                if(ST(TS,"mgfile")==0){
                        mgout=TRUE;
                        }
        	if(ST(TS,"maskfile")==0){
                	maskout=TRUE;
                	}
        	if(ST(TS,"set_psf")==0){
                	control>>temp_no;
                	control>>temp_string;
                	if(ST(TS,"meters")==0){
                        	Ax=Ay=temp_no;
                        	psf_set_meters=TRUE;
                        	}
                	if(ST(TS,"pixels")==0){
                        	ax=ay=(int)temp_no;
                        	psf_set_pixels=TRUE;
                        	}
                	}
	
        	if(ST(TS,"statfile")==0){
                	control>>statfile;
                	statfile_set=TRUE;
                	}
        	if(ST(TS,"newstart")==0){
                	newstart=TRUE;
                	}
        	if(ST(TS,"parameter")==0){
                	control>>variable;
                	}
        	if(ST(TS,"rank")==0){
                	control>>rank;
                	}
         	if(ST(TS,"dropoutlevel")==0){
                	control>>dol;
                	}
		}
        while(ST(TS,"endoptions")!=0&&i<MAX_OPTIONS&&!control.eof());
	checki(i);
	if(swap_floats==TRUE){
		cout<<"Byte swapping on\n";
		}
// setup FFT arrays and initilaize them
// allocate the working memory
	//ystride=N+1;i this is for cray only
	ystride=N;//for internal rotuines only
        //ntable=100+2*(N+N);
        //table=new float[ntable+1];
        //nwork=512*2*N;
        //work=new float[nwork+1];
        data = new complex<COMPLEX_TYPE>[(N+1)*(N+1)];
        if(data==NULL){
		cout<<"ERROR: unable to allocate memory .\n";
		cout<<" data array "<<N*(N+1)<<endl;
		exit(3);
		}
	if(N==0){
		cout<<"ERROR: recon size is not set"<<endl;
		exit(5);
		}
	cout<<"Initializing FFT workspace \n";
	//init_work(data,N,ystride);
	cout<<"done \n";
        if(statfile_set==TRUE){
                cout<< "Allocating stat memory "<<endl;
                mask=new int [N*ystride];
                hfg=new int[NBINS];
                hbg=new int[NBINS];
                if(mask==NULL||hfg==NULL||hbg==NULL){
                        cout<<"ERROR: unable to allocate mask/hfg/hbg memory"<<endl;
                        exit(3);
                        }
		}
//RECONrecon
//big loop now,
	for(i=0;i<no_files;i++){
		cout<<"Main loop, starting file "<<iname[i]<<"\n";

// now read file into fft array, transform in place proagate and detransform in plave
        cout << "Initializing simulation workspace...";
	for(temp=0;temp<N*ystride;temp++){
                                data[temp]=complex<COMPLEX_TYPE>(0,0);
                                }
	cout<<"done\n";
        cout << "loading "<<iname[i]<<" as a DS.CC style textfile...";
        switch(btype[i]){
                case PLANE:{
                        intext(iname[i],data,nx[i],ny[i],N,ystride,phaselevels[i]);
                        }
                        break;
                case GAUSSIAN:{
                        intext_gaussian(iname[i],data,nx[i],ny[i],N,ystride,phaselevels[i],psize[i]);
                        }
                        break;
                default:{
                        cout<<"Unknown beam type, probably internal error, assuming PLANE\n";
                        intext(iname[i],data,nx[i],ny[i],N,ystride,phaselevels[i]);
                        }
                }
	cout<<"done\n";
	if(R_curve[i]!=0){
		cout<<"adding wavefront to hologram...";
		add_wavefront(data,nx[i],ny[i],N,ystride,R_curve[i],psize[i],wavelength);
		cout<<"done\n";
		}
        if(hol_out==TRUE){
                cout << "Writing out hol file,constructing filename\n";
                sprintf(temp_string, "%s",iname[i]);
                output(complex_out,temp_string,data,N,ystride);
                }
	cout<<"Start FFT... ";
	fft(data,N,ystride,1);
	if(fftout==TRUE){
                sprintf(temp_string,"fft%s",iname[i],N,N);
		cout<<"outputing the complex<COMPLEX_TYPE> angular-spectrum before prop to "<<temp_string<<"\n";
                output(TRUE,temp_string,data,N,ystride);
       		}
	if(spat_filt==TRUE){
		filter(data,N,filter_width);
		}	

	cout<<"done"<<endl<<"Propagate field to "<<F[i]<<" meters... ";
	prop2d(data,F[i],psize[i],wavelength,N,ystride);
	cout<<"done"<<endl<<"Start iFFT... ";
	fft(data,N,ystride,-1);
	cout<<"done."<<endl;
	if(recon_out==TRUE){
		cout << "Writing out recon file,constructing filename\n";
                sprintf(temp_string, "%s",oname[i]);
		output(complex_out,temp_string,data,N,ystride);
		}
	if(statfile_set==TRUE){
		for(temp=0;temp<N*ystride;temp++){
			mask[temp]=BG_TOKEN;
			}
		for(temp=0;temp<NBINS;temp++){
			hfg[temp]=0;
			hbg[temp]=0;
			}
		//check mask array
		//PSF calculate the image psf
		cout<<"setting the psf"<<endl;
		if(psf_set_pixels==TRUE&&psf_set_meters==FALSE){
			Ax=psize[i]*ax;
			Ay=psize[i]*ay;
			psf_set_meters=TRUE;
			}
		if(psf_set_meters==FALSE&&psf_set_pixels==FALSE){
        		cout<<", calculating the psf in meters";
        		Ax=image_psf_size(F[i],nx[i]*psize[i],wavelength);
        		Ay=image_psf_size(F[i],ny[i]*psize[i],wavelength);
        		}
		//********IMPORTANT CHANGE from simulate 1.00 ds_t3d version (change of (0,0,0))
		offx=(int)((N/2));//-(nx[i]/(2)));
		offy=(int)((N/2));//-(ny[i]/(2)));
		scale=1/psize[i];
	
		//MASK-Construct the mask data
		cout<<"narrow version fg is 1 pixel wide\n";
		cout<<"drawing the image mask using image data from "<<posfile[i]<<"\n";
		cout<<"the psf function is "<<ax<<" by "<<ay<<" pixels"<<"\n";
		cout<<"the offset is "<<offx<<","<<offy<<"\n";
		cout<<"the scaling factor is "<<scale<<" pixels/meter\n";
		draw_mask(posfile[i],mask,Ax,Ay,offx,offy,scale,N,ystride);
		//output fg and bg files if filenames are given
		if(maskout==TRUE){
               	 		sprintf(temp_string,"mask%s",iname[i],N,N);
				cout<<"outputing the mask image to "<<temp_string<<"\n";
                        	outint_bin(temp_string,mask,N,ystride);
                		}
                if(pgout==TRUE){
                                //mask the foreground
                                sprintf(temp_string,"pg%s",iname[i],N,N);
                                cout<<"generating the peak image"<<"\n";
                                pground=new complex<COMPLEX_TYPE>[N*ystride];
                                mask_array(data,pground,mask,N,PEAK_TOKEN);
                                cout<<"outputing the foreground image to "<<temp_string<<"\n";
                                output(complex_out,temp_string,pground,N,ystride);
                                }
		if(mgout==TRUE){
                                //mask the middleground
                                sprintf(temp_string,"mg%s",iname[i],N,N);
                                cout<<"generating the foreground image"<<"\n";
                                fground=new complex<COMPLEX_TYPE>[N*ystride];
                                mask_array(data,fground,mask,N,MG_TOKEN);
                                cout<<"outputing the foreground image to "<<temp_string<<"\n";
                                output(complex_out,temp_string,fground,N,ystride);
                                }
		if(bgout==TRUE){
                		//mask the background
                                sprintf(temp_string,"bg%s",iname[i],N,N);
                		cout<<"generating the background image"<<"\n";
                		bground=new complex<COMPLEX_TYPE>[N*ystride];
                		mask_array(data,bground,mask,N,BG_TOKEN);
                		cout<<"outputing the background image to "<<temp_string<<"\n";
                		output(complex_out,temp_string,bground,N,ystride);
				}
		
		//STATS---------------
		cout<<"collecting stats "<<endl;
		collect_stats(data,mask,N,ystride,&pg_stat,&fg_stat,&mg_stat,&bg_stat);
		//calculate the mean
		pg_stat.xbar=pg_stat.sx/pg_stat.n;
		fg_stat.xbar=fg_stat.sx/fg_stat.n;
		mg_stat.xbar=mg_stat.sx/mg_stat.n;
		bg_stat.xbar=bg_stat.sx/bg_stat.n;
			
		//calculate the standard deviation
                pg_stat.sdx=sd(pg_stat.ssqx,pg_stat.sx,pg_stat.xbar,pg_stat.n);
		fg_stat.sdx=sd(fg_stat.ssqx,fg_stat.sx,fg_stat.xbar,fg_stat.n);
		mg_stat.sdx=sd(mg_stat.ssqx,mg_stat.sx,mg_stat.xbar,mg_stat.n);
		bg_stat.sdx=sd(bg_stat.ssqx,bg_stat.sx,bg_stat.xbar,bg_stat.n);
		// estimate the efficiency using the mid ground
		efficiency=100.0*mg_stat.xbar*mg_stat.n/(mg_stat.xbar*mg_stat.n+bg_stat.xbar*bg_stat.n);
		if(histogram==TRUE){
        		cout<<"collecting the histogram data"<<endl;
        		sprintf(temp_string,"%s.hist",iname[i]);
			collect_histogram(data,mask,N,hfg,hbg,&fg_stat,&bg_stat,temp_string,dol);
        		cout<<"the drop out number at level "<<dol<<" is "<<fg_stat.dropouts<<endl;
        		cout<<"this is a dropout percentage of "<<100*fg_stat.dropouts/fg_stat.n<<endl;
        		}


		if(linedump_on==TRUE){
			cout<<"dumping line "<<endl;
			sprintf(temp_string,"%s.ld",iname[i]);
			linedump(posfile[i],temp_string,data,offx,offy,N,ystride,scale);
			}	


		//estimate the mean contrast
		contrast=fg_stat.xbar/bg_stat.xbar;

		//print out a summary of results
		cout<<"stats collected\n";
		cout<<"total pixels  fg "<<fg_stat.n<<" mg "<<mg_stat.n<<" bg "<<bg_stat.n<<"\n";
		cout<<"average intensitys \n";
		cout<<"fg "<<fg_stat.xbar<<" mg "<<mg_stat.xbar<<" bg "<<bg_stat.xbar<<"\n";
		cout<<"average contrast ratio, fg:bg "<<contrast<<":1"<<endl;
		cout<<"foreground % variation "<<100.0*fg_stat.sdx/fg_stat.xbar<<"% \n";
		cout<<"standard deviations \n";
		cout<<"foreground "<<fg_stat.sdx<<" background "<<bg_stat.sdx<<"\n";
		cout<<"foreground min and max\n";
		cout<<"max "<<fg_stat.xmax<<" min "<<fg_stat.xmin<<"\n";
		cout<<"background max and min\n";
		cout<<"max "<<bg_stat.xmax<<" min "<<bg_stat.xmin<<endl;
		cout<<"estimate of efficiency (fg energy/total energy) "<<efficiency<<" % \n";
		
		};
	}//this should be the big loop end
return 0;
}






 
void draw_psf(int xs,int ys,int xf,int yf,int *data,int ,int ystride,int token){
int x,y;
for(x=xs;x<=xf;x++){
        for(y=ys;y<=yf;y++){
                if(data[x+ystride*y]<token)data[x+ystride*y]=token;
                }
        }
}
 
 
float image_psf_size(float focal_length,float hologram_aperture_size,float wavelength)
{
cout<<"setting the psf to the width for a sinc squared\n";
return wavelength*focal_length/hologram_aperture_size;
}
 
void mask_array(complex<COMPLEX_TYPE> *arrayin,complex<COMPLEX_TYPE> *arrayout,int *mask,int N,int key){
cout<<"warning mask array odd stride unsafe"<<endl;
int reg;
for(reg=0;reg<N*N;reg++){
        if(mask[reg]==key){
                arrayout[reg]=arrayin[reg];
                }
        else{
                arrayout[reg]=0;
                }
        }
}
 
 
 



 
void draw_mask(char *name,int *mask,float ax,float ay,int offx,int offy,float scale, int N,int ystride){
// nice new precise draw mask using floating point psf function
        ifstream is;
        float 	x,y,xs,xf,ys,yf,Offx,Offy,hx,hy;
	int	ixs,ixf,iys,iyf;
        int 	temp,i;
	
        read_ok(name);
        is.open(name);
        if(is.bad()){
                cout<<"error: unable to open image posistion data file "<<name<<"\n";
                exit(3);
                }
        i=-1;
	cout<<"setting halfmax width to 1/2.3 1st zero width (sinc squared)\n";
	hx=ax/2.3;
	hy=ay/2.3;
	Offx=offx;//+.5;	//to take later rounding into account
	Offy=offy+.5;
        do{
                i++;
                is>>x>>y;
		xs=x-ax;
		xf=x+ax;
		ys=y-ay;
		yf=y+ay;
		ixs=(int)((float)Offx+scale*xs);
		ixf=(int)((float)Offx+scale*xf);
		iys=(int)((float)Offy+scale*ys);
		iyf=(int)((float)Offy+scale*yf);
                if(!(temp=is.eof())){
                        //draw the image aperture
			//mg
                        if(     (ixs>=0)&&
                                (ixf<N)&&
                                (iys>=0)&&
                                (iyf<N)){
                                        draw_psf(ixs,iys,ixf,iyf,mask,N,ystride,MG_TOKEN);
                                        }
			else{
				cout<<"warning image off mask at "<<i<<"\n";
				}
			//fg
                xs=x-hx;
                xf=x+hx;
                ys=y-hy;
                yf=y+hy;
                ixs=(int)((float)Offx+scale*xs);
                ixf=(int)((float)Offx+scale*xf);
                iys=(int)((float)Offy+scale*ys);
                iyf=(int)((float)Offy+scale*yf);
                   if(     (ixs>=0)&&
                                (ixf<N)&&
                                (iys>=0)&&
                                (iyf<N)){
                                        draw_psf(ixs,iys,ixf,iyf,mask,N,ystride,FG_TOKEN);
                                        }
	//narrow central region 1 pixel peak (pg)
                ixs=(int)((float)Offx+scale*x);
                iys=(int)((float)Offy+scale*y);
			if(     (ixs>=0)&&
                                (ixs<N)&&
                                (iys>=0)&&
                                (iys<N)){
                                        mask[ixs+ystride*iys]=PEAK_TOKEN;
                                        }
                        }
		}
        while(!temp);
        cout<<i<<" image pos values read\n";
        is.close();
        return;
}
 
float intensity(complex<COMPLEX_TYPE> data){
float temp;
temp =  real(data) * real(data) + imag(data) * imag(data);
return temp;
}
 
 
 

void
collect_histogram(complex<COMPLEX_TYPE> *recon,
                int *mask,
                int N,
                int *hfg,
                int *hbg,
                STAT *fg,
                STAT *bg,
                char *histname,
                float dol){
float scale,max;
int     in,i,nodos=0;
ofstream        os;
dol=dol*NBINS;
max=fg->xmax;
if(bg->xmax>max){
        max=bg->xmax;
        }
cout<<max<<" ";
scale=(NBINS)/(max+SMALL_NO);
cout<<scale<<" dol "<<dol<<" nbins "<<NBINS<<endl;
for(i=0;i<N*N;i++){
        in=(int)((float)(intensity(recon[i])*scale));
        if(mask[i]>=FG_TOKEN){
                hfg[in]++;
                if(in<dol)nodos++;
                }
        else{
                hbg[in]++;
                }
        }
os.open(histname);
for(i=0;i<NBINS;i++){
        os<<i<<" "<<(float)i/scale<<" "<<hfg[i]<<" "<<hbg[i]<<"\n";
        }
os.close();
fg->dropouts=nodos;
}





void
se(int i)
{
        cout << "gen shapes: switch read error at position " << i << "\n";
}
 
int
find_switch(int number, char **list, char *comp)
{
        int             i, ret = -1;
        for (i = 0; i < number; i++) {
                if (list[i][0] == '-') {
                        //then it is a switch
                                if (strcmp(comp, list[i] + 1) == 0) {
                                //then it is the correct one
                                        ret = i;
                                i = number;
                        }
                }
        }
        return ret;
}
 
 
 
void read_ok(char *name){
        if(access(name,R_OK)!=0){
                cout<<"ERROR: unable to open file "<<name<<" exiting\n";
                cout.flush();
                cerr<<"ERROR: unable to open file "<<name<<" exiting\n";
                exit(3);
                }
}
void checki(int temp){
	if(temp==MAX_OPTIONS){
		cout<<"ERROR: read more than the maximum number of options\n";
		cout<<"       probably control file error, exiting"<<endl;
		exit(4);
		}
}
void init_work(complex<COMPLEX_TYPE> *data,int size,int ystride){
//	cray only fft(data, size, ystride,0);
}

void fft(complex<COMPLEX_TYPE> *data,int xsize,int ysize,int direction){
long oldt,newt;
unsigned long dim[4];
// this does a fft in place using a lut in table and workspace work
int incx,incy;
float scale;
/* ***cray only
//convert this to cray fortran function CFFT2D
//CFFT2D(int * isign,int *n1, int * n2,float * scale,complex<COMPLEX_TYPE> * data,int *xinc,in t* yinc,
//complex<COMPLEX_TYPE> * data, int * *xinc,int * yinc,float * trig_table,int * ntable,float * work_space,
//int * nwork);

        incy=ysize;
        incx=1;
        scale=1.0/(float)xsize;

        CFFT2D(&direction,&xsize,&xsize,&scale,
                data,&incx,&incy,
                data,&incx,&incy,
                table,&ntable,work,&nwork);
************cray only */



// using the built in fft routines..................
oldt=(long)time(0);
dim[0]=dim[1]=dim[2]=dim[3]=xsize;
fourn_c(data,dim,2,direction);
normalizefft(data,dim,2);
newt=(long)time(0);
cout<<" "<<(int)(newt-oldt)<<" s ";
}




void
prop2d(complex<COMPLEX_TYPE> * data, float distance,float pixelsize,float wavelength,int N,int ystride)
{
        int             x, y;
        long            pointer;
        register double factor1,factor2;
        int Ny;
        //fac = -(PI * distance * wavelength) / (/*4 */ pixelsize*pixelsize*N*N);
	factor1=2*PI*distance/wavelength;			//converts distance to phase
	factor2=wavelength*wavelength/(N*N*pixelsize*pixelsize);//converts x,y to fx,fy
        Ny=ystride;//odd ystride for performance
//cannot vectoise these loops as the call to arg etc in rot inhibits it despite atan2 being ok
        for (y = 0; y < N / 2; y++) {
                //1 st quadrent
                        for (pointer = Ny * y, x = 0; x < N / 2; x++, pointer++) {
                        data[pointer] = rot(data[pointer], fprop(x,y,factor1,factor2));
                }
        }
        for (y = N / 2; y >= 0; y--) {
                //3 rd quadrent
                        for (pointer = Ny * (N - y), x = 0; x < N / 2; x++, pointer++) {
                        data[pointer] = rot(data[pointer], fprop(x,y,factor1,factor2));
                }
        }
        for (y = 0; y < N / 2; y++) {
                //2 nd quadrent
                        for (pointer = (Ny * y + N / 2), x = N / 2-1; x >= 0; x--, pointer++) {
                        data[pointer] = rot(data[pointer], fprop(x,y,factor1,factor2));
                }
        }

        for (y = N / 2; y >= 0; y--) {
                //4 th quadrent
                        for (pointer = (Ny * (N - y)) + N / 2, x = N / 2-1; x >= 0; x--, pointer++) {
                        data[pointer] = rot(data[pointer], fprop(x,y,factor1,factor2));
                }
        }
}


inline complex<COMPLEX_TYPE>
rot(complex<COMPLEX_TYPE> temp, double phase)
{
        //rotates a complex<COMPLEX_TYPE> by phase radians
                return (polar(sqrt(norm(temp)), arg(temp) + phase));
}

// possible solution to vectroizing but doesn't work
inline complex<COMPLEX_TYPE> rotf(double r,double i,double phase){
complex<COMPLEX_TYPE> temp;
	temp=polar(sqrt(r*r+i*i),(atan2(r,i)+phase));
return temp;
}

void outint_txt(char *name,int *data,int n,int ys){
        ofstream os;
        int x,y;
       	os.open(name);
        for(y=0;y<n*(ys);y+=(ys)){
                for(x=0;x<n;x++){
			os<<data[(x+y)]<<"\n";
                }
        }
	os.close();
}
void
outcom_bin(char *name, complex<COMPLEX_TYPE> * data, int size,int ystride)

{
        FILE           *of;
        long            pointer;
        int x,y,i=0;
	float 	re,im;
	char	temp[MAX_S];
	while(name[i]!='\0'&&name[i]!='.'&&i<strlen(name)){
		temp[i]=name[i];
		i++;
		}
	temp[i]='\0';
	sprintf(name,"%s_cf_%d_%d.d",temp,size,size);
	cout<<"recon name is "<<name<<endl;
        of = fopen(name, "wb");
        for(y=0;y<size*(ystride);y+=(ystride)){
                for(x=0;x<size;x++){
                       pointer=x+y;
			re=real(*(data+pointer));
			im=imag(*(data+pointer));
                        fwrite(&re,sizeof(float),1,of);
                        fwrite(&im,sizeof(float),1,of);
		//	fprintf(of,"%f\n%f\n",real(data[pointer]),imag(data[pointer]));
                }
        }
        fclose(of);
}

void
outint_from_com(char *name, complex<COMPLEX_TYPE> * data, int size,int ystride)
{
        FILE           *of;
        long            pointer;
        int x,y,i=0;
        float   I;
        char    temp[MAX_S];
        while(name[i]!='\0'&&name[i]!='.'&&i<strlen(name)){
                temp[i]=name[i];
                i++;
                }
        temp[i]='\0';
        sprintf(name,"%s_f_%d_%d.d",temp,size,size);
        cout<<"recon name is "<<name<<endl;
        of = fopen(name, "wb");
        for(y=0;y<size*(ystride);y+=(ystride)){
                for(x=0;x<size;x++){
                       pointer=x+y;
                        I=intensity(*(data+pointer));
                       	//experimental alpha2sun conversion
			 if(swap_floats==TRUE){
				I=swap_f(I);
				}
			fwrite(&I,sizeof(float),1,of);
                //      fprintf(of,"%f\n%f\n",real(data[pointer]),imag(data[pointer]));
                }
        }
        fclose(of);
}

void
outint_bin(char *name, int * data, int size,int ystride)
{
        FILE           *of;
        long            pointer;
        int x,y,i=0;
        float   I;
        char    temp[MAX_S];
        while(name[i]!='\0'&&name[i]!='.'&&i<strlen(name)){
                temp[i]=name[i];
                i++;
                }
        temp[i]='\0';
        sprintf(name,"%s_i_%d_%d.d",temp,size,size);
        cout<<"writing integer "<<name<<endl;
        of = fopen(name, "wb");
        for(y=0;y<size*(ystride);y+=(ystride)){
                        fwrite(data+y,sizeof(int),size,of);
        }
        fclose(of);
}

void output(BOOL complex_out,char *name, complex<COMPLEX_TYPE> * data, int size,int ystride)
{
if(complex_out==TRUE){
	outcom_bin(name,data,size,ystride);
	}
	else{
	outint_from_com(name,data,size,ystride);
	}
}
 
void
collect_stats(complex<COMPLEX_TYPE>  *data,
                int    *mask,
                int     N,
		int	ys,
		STAT   *pg,
                STAT   *fg,
		STAT   *mg,
                STAT   *bg)
{
//finds the sum, the sum of the squares, the size of the samples etc,
//and the min and max values for the intensity
//for both the foreground and the background
 
        int     x,y,p;
        complex<COMPLEX_TYPE> temp;
 
//use local variables because not all compilers are capable of
//optimizing sums involving structures (esp. cray or cfront)
 
        float   I;
        float   sump=0,sumsqp=0,maxp=-1e6,minp=1e6;
        int     totp=0;
	float	sumf=0,sumsqf=0,maxf=-1e6,minf=1e6;
	int 	totf=0;
        float   summ=0,sumsqm=0,maxm=-1e6,minm=1e6;
        int     totm=0;
        float   sumb=0,sumsqb=0,maxb=-1e6,minb=1e6;
        int     totb=0;


 
        for(y=0;y<N;y++){
                for(x=0;x<N;x++){
                        p=x+y*ys;
			I=real(data[p])*real(data[p])+
                          imag(data[p])*imag(data[p]);
// first collect the stuff on the peak of the line,
                     if(mask[p]>=PEAK_TOKEN){
                                (totp)++;
                                (sump)+=I;
                                (sumsqp)+=(I*I);
                                if(I>maxp)maxp=I;
                                if(I<minp)minp=I;
                                }
		     if(mask[p]>=FG_TOKEN){
                                (totf)++;
                                (sumf)+=I;
                                (sumsqf)+=(I*I);
                                if(I>maxf)maxf=I;
                                if(I<minf)minf=I;
                                }
// then collect the stuff on the whole line width 
			if(mask[p]>=MG_TOKEN){
                                (totm)++;
                                (summ)+=I;
                                (sumsqm)+=(I*I);
                                if(I>maxm)maxm=I;
                                if(I<minm)minm=I;
                                }
// then collect the stuff on the background
			else{
				(totb)++;
                                (sumb)+=I;
                                (sumsqb)+=(I*I);
                                if(I>maxb)maxb=I;
                                if(I<minb)minb=I;
                                }
                        }
                }
        //transfer the local values to the structures
	
        pg->sx=sump;
        pg->ssqx=sumsqp;
        pg->n=totp;
        pg->xmax=maxp;
        pg->xmin=minp;
        fg->sx=sumf;
        fg->ssqx=sumsqf;
        fg->n=totf;
        fg->xmax=maxf;
        fg->xmin=minf;
        mg->sx=summ;
        mg->ssqx=sumsqm;
        mg->n=totm;
        mg->xmax=maxm;
        mg->xmin=minm;
        bg->sx=sumb;
        bg->ssqx=sumsqb;
        bg->n=totb;
        bg->xmax=maxb;
        bg->xmin=minb;
}

void intext_gaussian(char *iname,complex<COMPLEX_TYPE> *data, int xsize, int ysize ,int N,int ys,int phaselevels,float psize){
ifstream        is;
int             x,y,m,ph;
long            pointer;
float           h2,xf,yf,hx,hy,mag;
hx=xsize*psize/2;
hy=ysize*psize/2;
h2=hx*hx;
//magnitude of gaussian, 1/e width = 1./2 hologram size
        read_ok(iname);
        is.open(iname);
        if(is.bad()){
        cout<<"error could not open text file "<<iname<<"\n";
        exit(3);
        }
        xsize /= 2;
        ysize /= 2;
        m = N / 2;
        for (y = m - ysize,yf=-hy; y < m + ysize; y++,yf+=psize) {
                for (x = m - xsize,xf=-hx; x < m + xsize; x++,xf+=psize) {
                        pointer = x + y * (ys);
                        is>>ph;
                        mag=MAGCONST*exp(-((xf*xf+yf*yf)/(h2)));
                        data[pointer] = polar((double)mag,(2*PI*ph/phaselevels)*detune);
                }
        }
is.close();
}
 
 
 
void intext(char *iname,complex<COMPLEX_TYPE> *data, int xsize, int ysize ,int N,int ys,int phaselevels){
ifstream        is;
int             x,y,m,ph;
long            pointer;
        read_ok(iname);
	is.open(iname);
        if(is.bad()){
        cout<<"error could not open text file "<<iname<<"\n";
        exit(3);
        }
        xsize /= 2;
        ysize /= 2;
        m = N / 2;
        for (y = m - ysize; y < m + ysize; y++) {
                for (x = m - xsize; x < m + xsize; x++) {
                        pointer = x + y * (ys);
                        is>>ph;
                        if(ph!=-1){
			data[pointer] = polar(1.0,(2*PI*ph/phaselevels)*detune);
			}
                }
        }
is.close();
}
 
float sd(float ssqx,float sx,float xbar,int n){
return sqrt(((ssqx)+(n*xbar*xbar)-(2*xbar*sx))/n);
}

void add_wavefront(complex<COMPLEX_TYPE> *data, int xsize,int ysize, int N, int ys,float R_curve,float psize,float wavelength){
//adds a spherical wavefront to the hologram

	int	sign=1;//direction of curvature
	int m,x,y;
	long pointer;
	float	phase,fac1,fac2;
//only add wavefront to active part of hologram (to add to all xsize and ysize=N)
	cout<<"R_c="<<R_curve<<" ";
	if(R_curve<0)sign=-1;
        xsize /= 2;
        ysize /= 2;
        m = N / 2;
	fac1=2*PI*sign/wavelength;
	fac2=psize*psize;
	R_curve=R_curve*R_curve;
        for (y = - ysize; y <  ysize; y++) {
                for (x = - xsize; x <  xsize; x++) {
			//posistion the pointer correctly
                        pointer = (m+x) + (y+m) * (ys);
			phase=fac1*sqrt(R_curve+fac2*(x*x+y*y));
                        data[pointer] = rot(data[pointer],phase);
                }
        }
} 



void linedump(char *posfile,char *dumpfile,complex<COMPLEX_TYPE> *data,int offx,int offy,int size,int ystride,float scale){
ifstream is;
ofstream os;
float fx,fy,I;
int x,y,ox=-9999,oy=-9999,temp,i=0,p;
is.open(posfile);
os.open(dumpfile);
cout<<"writing line dump...";
do{
if(!(temp=is.eof())){
	is>>fx>>fy;
	x=(int)((float)offx+(scale*fx));
	y=(int)((float)offy+(scale*fy));
	if(x!=ox||y!=oy){
		i++;
		ox=x;
		oy=y;
		p=x+y*ystride;
		I=real(data[p])*real(data[p])+imag(data[p])*imag(data[p]);
		os<<I<<"\n";
		}
	}
}
while(!temp);
cout<<i<<"points done\n";
}


//************************numerical recipes fft routine modified by carl to include complex<COMPLEX_TYPE>


/*modified from NUMERICAL RECIPES in least way to take *complex<COMPLEX_TYPE> with range
/*from data[0] to data[N[1]*N[2]*...*N[no_dims-1]]
*/
#define SWAP(a,b) TEMP=(a);(a)=(b);(b)=TEMP
void fourn_c(complex<COMPLEX_TYPE> *data, unsigned long *nn, int ndim, int isign)
{
	int idim;
	unsigned long i1,i2,i3,i2rev,i3rev,ip1,ip2,ip3,ifp1,ifp2;
	unsigned long ibit,k1,k2,n,nprev,nrem,ntot;
	complex<COMPLEX_TYPE> TEMP,W,WP;
	double theta,wtemp;
	for (ntot=1,idim=1;idim<=ndim;idim++)
		ntot *= nn[idim];
	nprev=1;
	for (idim=ndim;idim>=1;idim--) {
		n=nn[idim];
		nrem=ntot/(n*nprev);
		ip1=nprev << 1;
		ip2=ip1*n;
		ip3=ip2*nrem;
		i2rev=1;
		for (i2=1;i2<=ip2;i2+=ip1) {
			if (i2 < i2rev) {
				for (i1=i2;i1<=i2+ip1-2;i1+=2) {
					for (i3=i1;i3<=ip3;i3+=ip2) {
						i3rev=i2rev+i3-i2;
						SWAP(data[(i3/2)],data[(i3rev/2)]);
					}
				}
			}
			ibit=ip2 >> 1;
			while (ibit >= ip1 && i2rev > ibit) {
				i2rev -= ibit;
				ibit >>= 1;
			}
			i2rev += ibit;
		}
		ifp1=ip1;
		while (ifp1 < ip2) {
			ifp2=ifp1 << 1;
			theta=isign*6.28318530717959/(ifp2/ip1);
			wtemp=sin(0.5*theta);
			WP=complex<COMPLEX_TYPE>(-2.0*wtemp*wtemp, sin(theta));
			W=complex<COMPLEX_TYPE>(1.0,0.0);
			for (i3=1;i3<=ifp1;i3+=ip1) {
				for (i1=i3;i1<=i3+ip1-2;i1+=2) {
					for (i2=i1;i2<=ip3;i2+=ifp2) {
						k1=i2;
						k2=k1+ifp1;
						TEMP=W*data[k2/2];
						data[k2/2]=data[k1/2]-TEMP;
						data[k1/2]=data[k1/2]+TEMP;
					}
				}
				W+=W*WP;
			}
			ifp1=ifp2;
		}
		nprev *= n;
	}

}
#undef SWAP

void normalizefft(complex<COMPLEX_TYPE> *data, unsigned long *nn, int ndim)
{
	long i,N;
	double scale;
	N=1;
	for(i=1;i<=ndim;i++){
		N*=nn[i];
	}
	scale=1.0/sqrt((double)N);
	for(i=0;i<N;i++){
		data[i]*=scale;
	}
	return;
}

void filter(complex<COMPLEX_TYPE> *data,int N,int w){
int 	w1,w2,w3;
int 	x,y;
long	pos;

w1=w+1;
w2=w;
w3=N-w;

for(y=0;y<N;y++){
	for(x=0;x<N;x++){
		if(	!(((x<w1)&&(y<w1))||
			  ((x<w2)&&(y>w3))||
			  ((x>w3)&&(y<w2))||
			  ((x>w3)&&(y>w3)))){
//this means we are !not in a keep zone so set to 0
			data[y*N+x]=complex<COMPLEX_TYPE>(0,0);
			}
		}
	}
}
				

